from .Entity import Entity


EXTRA_HEADER = """File automatically generated by pymoordyn"""


class Intro(Entity):
    """The introduction section
    """
    def __init__(self):
        """Constructor
        """
        Entity.__init__(self,
                        "MoorDyn Input File",
                        extra_header=EXTRA_HEADER)



import os
import tempfile
from .Intro import Intro
from .. import moordyn


INPUT_FNAME = "mooring.txt"


class Mooring:
    """A mooring system that can be programmatically generated

    The target of this class is not checking that the provided inputs are
    correct. MoorDyn itself will take care of that.
    """
    def __init__(self):
        # Generate the temporal folder where we are working
        self.__root_folder = tempfile.mkdtemp()
        self.__input_file_path = os.path.join(self.__root_folder, "moordyn.txt")
        # The lists of objects
        self.__intro = Intro()
        self.__line_materials = []
        self.__rod_materials = []
        self.__bodies = []
        self.__rods = []
        self.__connections = []
        self.__lines = []
        self.__failures = []
        self.__options = []
        # The MoorDyn instance
        self.__instance = None

    def AddLineMaterial(self, mat):
        self.__line_materials.append(mat)

    def AddRodMaterial(self, mat):
        self.__rod_materials.append(mat)

    def AddBody(self, body):
        self.__bodies.append(body)
        body.name = len(self.__bodies)

    def AddRod(self, rod):
        self.__rods.append(rod)
        rod.name = len(self.__rods)

    def AddConnection(self, con):
        self.__connections.append(con)
        con.name = len(self.__connections)

    def AddLine(self, line):
        self.__lines.append(line)
        line.name = len(self.__lines)

    def AddFailure(self, failure):
        self.__failures.append(failure)

    def AddOption(self, option):
        self.__options.append(option)

    def __write_input_file(self):
        with open(self.__input_file_path, "w") as f:
            f.write(self.__intro.get_header())
            self.__write_section(f, self.__line_materials)
            self.__write_section(f, self.__rod_materials)
            self.__write_section(f, self.__bodies)
            self.__write_section(f, self.__rods)
            self.__write_section(f, self.__connections)
            self.__write_section(f, self.__lines)
            self.__write_section(f, self.__failures)
            self.__write_section(f, self.__options)
            f.write(self.__get_footer())

    def __write_section(self, f, lst):
        if not len(lst):
            return
        w = lst[0].get_field_widths()
        for v in lst:
            v.set_field_widths(w)
            w = v.get_field_widths()
        for v in lst:
            v.set_field_widths(w)
        hdr = False
        for v in lst:
            if not hdr:
                f.write(v.get_header())
                hdr = True
            f.write(v.get_values())

    def __get_footer(self):
        return ("-" * 80) + "\n"

    def Create(self):
        self.__write_input_file()
        print("Created the MoorDyn file: '{}'".format(self.__input_file_path))
        print("")
        print("=" * 80)
        print("")
        with open(self.__input_file_path, "r") as f:
            print(f.read())
        print("")
        print("=" * 80)
        print("")
        return moordyn.Create(filepath=self.__input_file_path)



from .Entity import Entity, PropsList
from .Line import LineConnection


class Failure(Entity):
    """A failure criteria
    """
    def __init__(self, con, lines, time, ten):
        """Constructor

        Parameters
        ----------
        con (LineConnection): The Line connection
        lines (list): The lines affected
        time (float): The failure tme delay
        ten (float): The failure tension
        """

        Entity.__init__(
            self,
            "FAILURE",
            field_names=["Attach", "Lines", "FailTime", "FailTen"],
            field_units=["(#)", "(#)", "(s)", "(N)"])
        self.__con = con
        self.__lines = lines
        self.__time = time
        self.__ten = ten
        self.__set_values()

    @property
    def con(self):
        return self.__con

    @con.setter
    def con(self, connection):
        self.__con = connection
        self.__set_values()

    @property
    def lines(self):
        return self.__lines

    @lines.setter
    def lines(self, lines):
        self.__lines = lines
        self.__set_values()

    @property
    def time(self):
        return self.__time

    @time.setter
    def time(self, v):
        self.__time = v
        self.__set_values()

    @property
    def ten(self):
        return self.__ten

    @ten.setter
    def ten(self,ten):
        self.__ten = ten
        self.__set_values()

    def __set_values(self):
        l = PropsList([line.name for line in self.__lines], ",")
        self.set_values([self.__con,
                         PropsList([line.name for line in self.__lines], ","),
                         self.__time,
                         self.__ten])



from .Entity import Entity, PropsList


class BodyConnection():
    """The body connection
    """
    def __init__(self):
        self.__fixed = True
        self.__coupled = False

    @property
    def fixed(self):
        return self.__fixed

    @fixed.setter
    def fixed(self, v):
        self.__fixed = v
        if v:
            self.__coupled = False

    @property
    def coupled(self):
        return self.__coupled

    @coupled.setter
    def coupled(self, v):
        self.__coupled = v
        if v:
            self.__fixed = False

    def __str__(self):
        """Get the string

        Returns
        -------
        str: The connection string
        """
        if self.__fixed:
            return "FIXED"
        if not self.__coupled:
            return "FREE"
        return "COUPLED"


class Body(Entity):
    """A body
    """
    def __init__(self, t, p, r, m, cog, inertia, cda, v):
        """Constructor

        Parameters
        ----------
        t (BodyConnection): The type of body connection
        p (list): Point (3 components)
        r (list): Rotation (3 components, radians)
        m (float): Node mass in the case of clump weights (kg)
        cog (float/list): The center of gravity (3 components). If just a single
                          number is provided, then (0, 0, cog) is considered
        v (float): Node displacement in the case of floats (m^3)
        inertia (float/list): The inertia (3 components). If just a single
                              number is provided, then (inertia, inertia, inertia)
                              is considered
        cda (float/list): Drag coefficients (3 components). If just a single
                          number is provided, then (cda, cda, cda) is considered
        ca (float/list): Added mass coefficients (3 components). If just a single
                         number is provided, then (ca, ca, ca) is considered
        """
        Entity.__init__(
            self,
            "BODIES",
            field_names=["ID", "Type", "X   Y   Z  ", "Roll  Pitch Yaw", "Mass", "CoG", "Volume", "Inertia", "CdA", "Ca"],
            field_units=["(#)", "(name)", "(m) (m) (m)", "(rad) (rad) (rad)", "(kg)", "(m)", "(m^3)", "(kg*m^2)", "(m^2)", "(-)"])
        self.__name = 0
        self.__t = t
        self.__p = PropsList(p, " ")
        self.__r = PropsList(r, " ")
        self.__m = m
        self.__cog = PropsList(cog, "|")
        self.__v = v
        self.__I = PropsList(inertia, "|")
        self.__cda = PropsList(cda, "|")
        self.__ca = PropsList(ca, "|")
        self.__set_values()

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n
        self.__set_values()

    @property
    def type(self):
        return self.__t

    @type.setter
    def type(self, t):
        self.__t = t
        self.__set_values()

    @property
    def p(self):
        return self.__p

    @p.setter
    def p(self, p):
        self.__p = PropsList(p, " ")
        self.__set_values()

    @property
    def r(self):
        return self.__r

    @r.setter
    def r(self, r):
        self.__r = PropsList(r, " ")
        self.__set_values()

    @property
    def mass(self):
        return self.__m

    @mass.setter
    def mass(self, v):
        self.__m = v
        self.__set_values()

    @property
    def cog(self):
        return self.__cog

    @cog.setter
    def cog(self, v):
        self.__cog = PropsList(v, "|")
        self.__set_values()

    @property
    def vol(self):
        return self.__v

    @vol.setter
    def vol(self, v):
        self.__v = v
        self.__set_values()

    @property
    def I(self):
        return self.__I

    @I.setter
    def I(self, v):
        self.__I = PropsList(v, "|")
        self.__set_values()

    @property
    def cda(self):
        return self.__cda

    @cda.setter
    def cda(self, v):
        self.__cda = PropsList(v, "|")
        self.__set_values()

    @property
    def ca(self):
        return self.__ca

    @ca.setter
    def ca(self, v):
        self.__ca = PropsList(v, "|")
        self.__set_values()

    def __set_values(self):
        self.set_values([self.__name,
                         self.__t,
                         self.__p,
                         self.__r,
                         self.__m,
                         self.__cog,
                         self.__v,
                         self.__I,
                         self.__cda,
                         self.__ca])



from .Entity import Entity, PropsList
from .Body import Body


class Connection(Entity):
    """A connection
    """
    def __init__(self, t, p, m, v, cda, ca):
        """Constructor

        Parameters
        ----------
        t (str): One of ("FIXED", "COUPLED", "FREE"), or a body
        p (list): Point (3 components)
        m (float): Node mass in the case of clump weights (kg)
        v (float): Node displacement in the case of floats (m^3)
        cda (float): Product of drag coefficient and projected area (assumed
                     constant in all directions) to calculate a drag force for
                     the node (m^2)
        ca (float): Added mass coefficient used along with V to calculate added
                    mass on node
        """
        Entity.__init__(
            self,
            "CONNECTION PROPERTIES",
            field_names=["ID", "Type", "X   Y   Z", "Mass", "Volume", "CdA", "Ca"],
            field_units=["(#)", "(name)", "(m) (m) (m)", "(kg)", "(m^3)", "(m^2)", "(-)"])
        self.__name = 0
        self.__t = t
        self.__p = PropsList(p, " ")
        self.__m = m
        self.__v = v
        self.__cda = cda
        self.__ca = ca
        self.__set_values()

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n
        self.__set_values()

    @property
    def type(self):
        return self.__t

    @type.setter
    def type(self, t):
        self.__t = t
        self.__set_values()

    @property
    def p(self):
        return self.__p

    @p.setter
    def p(self, p):
        self.__p = PropsList(p, " ")
        self.__set_values()

    @property
    def mass(self):
        return self.__m

    @mass.setter
    def mass(self, v):
        self.__m = v
        self.__set_values()

    @property
    def vol(self):
        return self.__v

    @vol.setter
    def vol(self, v):
        self.__v = v
        self.__set_values()

    @property
    def cda(self):
        return self.__cda

    @cda.setter
    def cda(self, v):
        self.__cda = v
        self.__set_values()

    @property
    def ca(self):
        return self.__ca

    @ca.setter
    def ca(self, v):
        self.__ca = v
        self.__set_values()

    def __set_values(self):
        self.set_values([self.__name,
                         self.__t,
                         self.__p,
                         self.__m,
                         self.__v,
                         self.__cda,
                         self.__ca])



from .Entity import Entity, PropsList


class RodMaterial(Entity):
    """A rod material
    """
    def __init__(self, name, d, w, cdn, can, cdt, cat):
        """Constructor

        Parameters
        ----------
        name (str): The name of the rod material
        d (float): The rod diameter
        w (float): The weight per meter of rod
        cdn (float): transverse drag coefficient (with respect to frontal area, d*l)
        can (float): transverse added mass coefficient (with respect to line displacement)
        cdt (float): tangential drag coefficient (with respect to surface area, pi*d*l)
        cat (float): tangential added mass coefficient (with respect to line displacement)
        """

        Entity.__init__(
            self,
            "ROD TYPES",
            field_names=["Name", "Diam", "Mass/m", "Cd", "Ca", "CdAx", "CaAx"],
            field_units=["(name)", "(m)", "(kg/m)", "(-)", "(-)", "(-)", "(-)"])
        self.__name = name
        self.__d = d
        self.__w = w
        self.__cdn = cdn
        self.__can = can
        self.__cdt = cdt
        self.__cat = cat
        self.__set_values()

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n
        self.__set_values()

    @property
    def d(self):
        return self.__d

    @d.setter
    def d(self, diam):
        self.__d = diam
        self.__set_values()

    @property
    def w(self):
        return self.__w

    @w.setter
    def w(self, weight):
        self.__w = weight
        self.__set_values()

    @property
    def cdn(self):
        return self.__cdn

    @cdn.setter
    def cdn(self, c):
        self.__cdn = c
        self.__set_values()

    @property
    def can(self):
        return self.__can

    @can.setter
    def can(self, c):
        self.__can = c
        self.__set_values()

    @property
    def cdt(self):
        return self.__cdt

    @cdt.setter
    def cdt(self, c):
        self.__cdt = c
        self.__set_values()

    @property
    def cat(self):
        return self.__cat

    @cat.setter
    def cat(self, c):
        self.__cat = c
        self.__set_values()

    def __set_values(self):
        self.set_values([self.__name,
                         self.__d,
                         self.__w,
                         self.__cdn,
                         self.__can,
                         self.__cdt,
                         self.__cat])


class RodConnection():
    """The rod connection
    """
    def __init__(self):
        self.__fixed = True
        self.__pinned = False
        self.__coupled = False
        self.__body = None

    @property
    def fixed(self):
        return self.__fixed

    @fixed.setter
    def fixed(self, v):
        self.__fixed = v
        if v:
            self.__coupled = False

    @property
    def pinned(self):
        return self.__pinned

    @pinned.setter
    def pinned(self, v):
        self.__pinned = v

    @property
    def coupled(self):
        return self.__coupled

    @coupled.setter
    def coupled(self, v):
        self.__coupled = v
        if v:
            self.__fixed = False

    @property
    def body(self):
        return self.__body

    @body.setter
    def body(self, v):
         self.__body= v
         if v:
             self.__coupled = False
             self.__fixed = True

    def __str__(self):
        """Get the string

        Returns
        -------
        str: The connection string
        """
        if self.__body:
            value = "BODY" + str(self.__body.name)
            if self.__pinned:
                value = value + "PINNED"
            return value
        if self.__fixed:
            if self.__pinned:
                return "PINNED"
            return "FIXED"
        if not self.__coupled:
            return "FREE"
        if self.__pinned:
            return "CPLDPIN"
        return "COUPLED"


class Rod(Entity):
    """A rod
    """
    def __init__(self, material, connection, p0, p1, n):
        """Constructor

        Parameters
        ----------
        material (RodMaterial): The rod material
        connection (RodConnection): The rod connection
        p0 (list): First point (3 components)
        p1 (list): Last point (3 components)
        n (int): Number of segments
        """

        Entity.__init__(
            self,
            "RODS",
            field_names=["ID", "RodType", "Type", "X0  Y0  Z0", "X1  Y1  Z1", "NumSegs", "Outputs"],
            field_units=["(#)", "(name)", "(-)", "(m) (m) (m)", "(m) (m) (m)", "(-)", "(-)"])
        self.__name = 0
        self.__mat = material
        self.__con = connection
        self.__p0 = PropsList(p0, " ")
        self.__p1 = PropsList(p1, " ")
        self.__n = n
        self.__set_values()

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n
        self.__set_values()

    @property
    def material(self):
        return self.__mat

    @material.setter
    def d(self, material):
        self.__mat = material
        self.__set_values()

    @property
    def connection(self):
        return self.__connection

    @connection.setter
    def connection(self, connection):
        self.__con = connection
        self.__set_values()

    @property
    def p0(self):
        return self.__p0

    @p0.setter
    def p0(self, p):
        self.__p0 = PropsList(p, " ")
        self.__set_values()

    @property
    def p1(self):
        return self.__p1

    @p1.setter
    def p1(self, p):
        self.__p1 = PropsList(p, " ")
        self.__set_values()

    @property
    def n(self):
        return self.__n

    @n.setter
    def n(self, n):
        self.__n = n
        self.__set_values()

    def __set_values(self):
        self.set_values([self.__name,
                         self.__mat.name,
                         self.__con,
                         self.__p0,
                         self.__p1,
                         self.__n,
                         "-"])



from .Entity import Entity


class Option(Entity):
    """An option
    """
    def __init__(self, name, value):
        """Constructor

        Parameters
        ----------
        name (str): The option name
        value (number): The option value
        """
        Entity.__init__(
            self,
            "OPTIONS")
        self.__name = name
        self.__value = value
        self.__set_values()

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n
        self.__set_values()

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value):
        self.__value = value
        self.__set_values()

    def __set_values(self):
        self.set_values([self.__value, self.__name])



from .Entity import Entity
from .Rod import Rod


class LineMaterial(Entity):
    """A line material
    """
    def __init__(self, name, d, w, ea, ba, ei, cdn, can, cdt, cat):
        """Constructor

        Parameters
        ----------
        name (str): The name of the line material
        d (float): The line diameter
        w (float): The weight per meter of line
        ea (float): The line stiffness, product of elasticity modulus and
                    cross-sectional area
        ba (float): The line internal damping (measured in N-s) or, if a
                    negative value is entered, the desired damping ratio (in
                    fraction of critical) for the line type
        ei (float): The bending stiffness
        cdn (float): transverse drag coefficient (with respect to frontal area, d*l)
        can (float): transverse added mass coefficient (with respect to line displacement)
        cdt (float): tangential drag coefficient (with respect to surface area, pi*d*l)
        cat (float): tangential added mass coefficient (with respect to line displacement)
        """

        Entity.__init__(
            self,
            "LINE TYPES",
            field_names=["Name", "Diam", "Mass/m", "EA", "BA/-zeta", "EI", "Cd", "Ca", "CdAx", "CaAx"],
            field_units=["(name)", "(m)", "(kg/m)", "(N)", "(N-s/-)", "(N-m^2)", "(-)", "(-)", "(-)", "(-)"])
        self.__name = name
        self.__d = d
        self.__w = w
        self.__ea = ea
        self.__ba = ba
        self.__ei = ei
        self.__cdn = cdn
        self.__can = can
        self.__cdt = cdt
        self.__cat = cat
        self.__set_values()

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n
        self.__set_values()

    @property
    def d(self):
        return self.__d

    @d.setter
    def d(self, diam):
        self.__d = diam
        self.__set_values()

    @property
    def w(self):
        return self.__w

    @w.setter
    def w(self, weight):
        self.__w = weight
        self.__set_values()

    @property
    def ea(self):
        return self.__ea

    @ea.setter
    def ea(self, v):
        self.__ea = v
        self.__set_values()

    @property
    def ba(self):
        return self.__ba

    @ba.setter
    def ba(self, v):
        self.__ba = v
        self.__set_values()

    @property
    def ei(self):
        return self.__ei

    @ei.setter
    def ei(self, v):
        self.__ei = v
        self.__set_values()

    @property
    def cdn(self):
        return self.__cdn

    @cdn.setter
    def cdn(self, c):
        self.__cdn = c
        self.__set_values()

    @property
    def can(self):
        return self.__can

    @can.setter
    def can(self, c):
        self.__can = c
        self.__set_values()

    @property
    def cdt(self):
        return self.__cdt

    @cdt.setter
    def cdt(self, c):
        self.__cdt = c
        self.__set_values()

    @property
    def cat(self):
        return self.__cat

    @cat.setter
    def cat(self, c):
        self.__cat = c
        self.__set_values()

    def __set_values(self):
        self.set_values([self.__name,
                         self.__d,
                         self.__w,
                         self.__ea,
                         self.__ba,
                         self.__ei,
                         self.__cdn,
                         self.__can,
                         self.__cdt,
                         self.__cat])


class LineConnection():
    """The line connection
    """
    def __init__(self, obj, end_point="A"):
        self.__obj = obj
        self.__end_point = end_point  # Used just with Rods

    @property
    def obj(self):
        return self.__obj

    @obj.setter
    def obj(self, v):
        self.__obj = v

    @property
    def end_point(self):
        return self.__end_point

    @end_point.setter
    def end_point(self, v):
        self.__end_point = v

    def __str__(self):
        """Get the string

        Returns
        -------
        str: The connection string
        """
        if isinstance(self.__obj, Rod):
            return "R" + str(self.__obj.name)
        return str(self.__obj.name)


class Line(Entity):
    """A line
    """
    def __init__(self, material, con0, con1, l, n):
        """Constructor

        Parameters
        ----------
        material (LineMaterial): The line material
        con0 (LineConnection): The line connection
        con1 (LineConnection): The line connection
        l (float): Unstretched length
        n (int): Number of segments
        """

        Entity.__init__(
            self,
            "LINES",
            field_names=["ID", "LineType", "AttachA", "AttachB", "UnstrLen", "NumSegs", "Outputs"],
            field_units=["(#)", "(name)", "(#)", "(#)", "(m)", "(-)", "(-)"])
        self.__name = 0
        self.__mat = material
        self.__con0 = con0
        self.__con1 = con1
        self.__l = l
        self.__n = n
        self.__set_values()

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, n):
        self.__name = n
        self.__set_values()

    @property
    def material(self):
        return self.__mat

    @material.setter
    def material(self, material):
        self.__mat = material
        self.__set_values()

    @property
    def con0(self):
        return self.__con0

    @con0.setter
    def con0(self, connection):
        self.__con0 = connection
        self.__set_values()

    @property
    def con1(self):
        return self.__con1

    @con1.setter
    def con1(self, connection):
        self.__con1 = connection
        self.__set_values()

    @property
    def l(self):
        return self.__l

    @l.setter
    def l(self, v):
        self.__l = v
        self.__set_values()

    @property
    def n(self):
        return self.__n

    @n.setter
    def n(self,n):
        self.__n = n
        self.__set_values()

    def __set_values(self):
        self.set_values([self.__name,
                         self.__mat.name,
                         self.__con0,
                         self.__con1,
                         self.__l,
                         self.__n,
                         "-"])



"""moordyn is a Python wrapper for the MoorDyn v2 C library

It is used in the same way than the regular MoorDyn C API:

.. code-block:: python

    import moordyn
    system = moordyn.Create(filepath="Mooring/lines.txt")
    moordyn.SetVerbosity(system, moordyn.LEVEL_MSG)
    moordyn.SetLogFile(system, "Mooring/lines.log")
    moordyn.SetLogLevel(system, moordyn.LEVEL_MSG)
    moordyn.Log(system, "We are ready to work!")
    x = [5.2, 0.0, -70.0,
         -2.6, 4.5, -70.0,
         -2.6, -4.5, -70.0]
    v = [0, ] * 9
    moordyn.Init(system, x, v)
    dt = 0.1
    moordyn.Step(system, x, v, 0.0, dt)
    moordyn.Close(system)

The majority of functions are returning error codes, exactly the same way the
MoorDyn v2 C API is doing
"""

from .moordyn import *
from . import Generator



"""
Copyright (c) 2022, Jose Luis Cercos-Pita <jlc@core-marine.com>

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

LEVEL_NONE = 4096
LEVEL_ERR = 3
LEVEL_WRN = 2
LEVEL_MSG = 1
LEVEL_DBG = 0


ERRCODE_SUCCESS = 0
ERRCODE_INVALID_INPUT_FILE = -1
ERRCODE_INVALID_OUTPUT_FILE = -2
ERRCODE_INVALID_INPUT = -3
ERRCODE_NAN_ERROR = -4
ERRCODE_MEM_ERROR = -5
ERRCODE_INVALID_VALUE = -6
ERRCODE_NON_IMPLEMENTED = -7
ERRCODE_UNHANDLED_ERROR = -255


BODY_TYPE_COUPLED = CONN_TYPE_COUPLED = -1
BODY_TYPE_FREE = CONN_TYPE_FREE = 0
BODY_TYPE_FIXED = CONN_TYPE_FIXED = 1
ROD_TYPE_COUPLED = -2
ROD_TYPE_CPLDPIN = -1
ROD_TYPE_FREE = 0
ROD_TYPE_PINNED = 1
ROD_TYPE_FIXED = 2


ENDPOINT_A = ENDPOINT_BOTTOM = 0
ENDPOINT_B = ENDPOINT_TOP = 1


#                                  MoorDyn2.h
#  =============================================================================

def Create(filepath=""):
    """Creates the MoorDyn system, loading the MoorDyn input file

    Keyword arguments:
    filepath (str): The file to load. If "" is passed, then "Mooring/lines.txt"
                    will be considered

    Returns:
    cmoordyn.MoorDyn: The MoorDyn instance
    """
    import cmoordyn
    return cmoordyn.create(filepath)


def NCoupledDOF(instance):
    """Get the number of coupled Degrees Of Freedom (DOF)

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: The number of DOFs
    """
    import cmoordyn
    return cmoordyn.n_coupled_dof(instance)


def SetVerbosity(instance, verbosity):
    """Set the verbosity level of the MoorDyn instance

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    verbosity (int): The Verbosity level. One of:
                     - moordyn.LEVEL_NONE
                     - moordyn.LEVEL_ERR
                     - moordyn.LEVEL_WRN
                     - moordyn.LEVEL_MSG
                     - moordyn.LEVEL_DBG

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.set_verbosity(instance, verbosity)


def SetLogFile(instance, filepath):
    """Set the verbosity level of the MoorDyn instance

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    filepath (str): The log file path

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.set_logfile(instance, filepath)


def SetLogLevel(instance, verbosity):
    """Set the log file verbosity level of the MoorDyn instance

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    verbosity (int): The Verbosity level. One of:
                     - moordyn.LEVEL_NONE
                     - moordyn.LEVEL_ERR
                     - moordyn.LEVEL_WRN
                     - moordyn.LEVEL_MSG
                     - moordyn.LEVEL_DBG

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.set_loglevel(instance, verbosity)


def Log(instance, level, msg):
    """Set the log file verbosity level of the MoorDyn instance

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    level (int): The Verbosity level. One of:
                     - moordyn.LEVEL_NONE
                     - moordyn.LEVEL_ERR
                     - moordyn.LEVEL_WRN
                     - moordyn.LEVEL_MSG
                     - moordyn.LEVEL_DBG
    msg (str): The message to log

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    import inspect

    head = "DBG "
    if level >= LEVEL_ERR:
        head = "ERR "
    elif level == LEVEL_WRN:
        head = "WRN "
    elif level == LEVEL_MSG:
        head = "MSG "
    caller = inspect.stack()[1]
    head = head + "{}:{} {}() : ".format(caller.filename,
                                         caller.lineno,
                                         caller.function)
    return cmoordyn.log(instance, level, head + msg)


def Init(instance, x, v):
    """Initializes the MoorDyn sytem, calculating initial conditions based on
    connection positions specified by array x and their velocities in array v

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Keyword arguments:
    x (list): Position of the coupled connections
    v (list): Velocity of the coupled connections

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.init(instance, list(x), list(v), 0)


def Init_NoIC(instance, x, v):
    """Initializes the MoorDyn sytem, without calculating initial conditions

    Use this when you are planning to call moordyn.Load() afterwards

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Keyword arguments:
    x (list): Position of the coupled connections
    v (list): Velocity of the coupled connections

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.init(instance, list(x), list(v), 1)


def Step(instance, x, v, t, dt):
    """Compute a time step

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    x (list): Position of the coupled connections
    v (list): Velocity of the coupled connections
    t (float): The time instant
    dt (float): The time step

    Returns:
    list: The forces acting on the coupled connections
    """
    import cmoordyn
    return cmoordyn.step(instance, list(x), list(v), t, dt)


def Close(instance):
    """This function deallocates the variables used by MoorDyn

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.close(instance)


def GetWaves(instance):
    """Get the wave kinematics instance

    The wave kinematics instance is only useful if WaveKin option is set to >= 2
    in the input file.

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    cmoordyn.MoorDynWaves: The waves manager
    """
    import cmoordyn
    return cmoordyn.get_waves(instance)


def ExternalWaveKinInit(instance):
    """Initializes the external Wave kinetics

    This is useless unless WaveKin option is set to 1 in the input file. If
    that is the case, remember calling this function after moordyn.Init()

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.ext_wave_init(instance)


def ExternalWaveKinGetN(instance):
    """Get the number of points where the waves kinematics shall be provided

    This is useless unless WaveKin option is set to 1 in the input file.

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: The number of points
    """
    import cmoordyn
    return cmoordyn.ext_wave_init(instance)


def GetWaveKinCoordinates(instance):
    """Get the points where the waves kinematics shall be provided

    The kinematics on those points shall be provided just if WaveKin is set
    to 1 in the input file

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    list (n, 3): The list of points
    """
    import cmoordyn

    coords = cmoordyn.ext_wave_coords(instance)
    n = len(coords) // 3
    points = []
    for i in range(n):
        points.append([coords[3 * i], coords[3 * i + 1], coords[3 * i + 2]])
    return points


def SetWaveKin(instance, u, a, t):
    """Set the kinematics of the waves

    Use this function if WaveKin option is set to 1 in the input file

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    u (list (n, 3)): The list of velocities evaluated in the points provided
                     by moordyn.GetWaveKinCoordinates()
    a (list (n, 3)): The list of accelerations evaluated in the points provided
                     by moordyn.GetWaveKinCoordinates()
    t (float): The simulation time

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn

    assert len(u) == len(a)
    uu, aa = [], []
    for i in range(len(u)):
        assert len(u[i]) == 3 and len(a[i]) == 3
        uu += u[i]
        aa += a[i]
    return cmoordyn.ext_wave_set(instance, uu, aa, t)


def GetNumberBodies(instance):
    """Get the number of bodies

    Remember that the first body index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: The number of bodies
    """
    import cmoordyn
    return cmoordyn.get_number_bodies(instance)


def GetBody(instance, body):
    """ Get a rigid body

    Remember that the first body index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    body (int): The rigid body index, starting at 1

    Returns:
    cmoordyn.MoorDynBody: The rigid body instance
    """
    import cmoordyn
    return cmoordyn.get_body(instance, body)


def GetNumberRods(instance):
    """Get the number rods

    Remember that the rod index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: The number of rods
    """
    import cmoordyn
    return cmoordyn.get_number_rods(instance)


def GetRod(instance, rod):
    """ Get a rod

    Remember that the first rod index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    rod (int): The rod index, starting at 1

    Returns:
    cmoordyn.MoorDynRod: The rod instance
    """
    import cmoordyn
    return cmoordyn.get_rod(instance, rod)


def GetNumberConnections(instance):
    """Get the number of connections

    Remember that the first connection index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: The number of connections
    """
    import cmoordyn
    return cmoordyn.get_number_connections(instance)


def GetConnection(instance, conn):
    """ Get a connection

    Remember that the first connection index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    conn (int): The connection index, starting at 1

    Returns:
    cmoordyn.MoorDynConnection: The connection instance
    """
    import cmoordyn
    return cmoordyn.get_connection(instance, conn)


def GetNumberLines(instance):
    """Get the number of mooring lines

    Remember that the first line index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: The number of mooring lines
    """
    import cmoordyn
    return cmoordyn.get_number_lines(instance)


def GetLine(instance, line):
    """ Get a mooring line

    Remember that the first line index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    line (int): The line index, starting at 1

    Returns:
    cmoordyn.MoorDynLine: The line instance
    """
    import cmoordyn
    return cmoordyn.get_line(instance, line)


def GetFASTtens(instance, n_lines):
    """Get the horizontal and vertical components of the fairlead and anchor
    tensions

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    n_lines (int): The number of lines

    Returns:
    list: The horizontal fairlead tension at each line
    list: The vertical fairlead tension at each line
    list: The horizontal anchor tension at each line
    list: The vertical anchor tension at each line
    """
    import cmoordyn
    data = cmoordyn.get_fast_tens(instance, n_lines)
    return data[0], data[1], data[2], data[3]


def Save(instance, filepath):
    """Save the MoorDyn system into a file that can be loaded afterwards to
    resume the simulation

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    filepath (str): The file path

    Returns:
    None
    """
    import cmoordyn
    cmoordyn.save(instance, filepath)


def Load(instance, filepath):
    """Load a previously saved MoorDyn system, in order to resume the simulation

    You still must read the same definition file with moordyn.Create() and call
    moordyn.Init_NoIC()

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    filepath (str): The file path

    Returns:
    None
    """
    import cmoordyn
    cmoordyn.load(instance, filepath)

#                                  Waves.h
#  =============================================================================


def GetWavesKin(instance, x, y, z, t):
    """ Get the wave kinematics

    Parameters:
    instance (cmoordyn.MoorDynWaves): The waves instance
    x (float): The x coordinate
    y (float): The y coordinate
    z (float): The z coordinate
    t (float): The simulation time

    Returns:
    u: The velocity (a tuple with 3 components)
    ud: The acceleration (a tuple with 3 components)
    zeta: The wave height
    pdyn: The dynamic pressure
    """
    import cmoordyn
    return cmoordyn.waves_getkin(instance, x, y, z, t)


#                                  Body.h
#  =============================================================================


def GetBodyID(instance):
    """ Get the body id

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance

    Returns:
    i: The body id
    """
    import cmoordyn
    return cmoordyn.body_get_id(instance)


def GetBodyType(instance):
    """ Get the body type

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance

    Returns:
    t: The body type
    """
    import cmoordyn
    return cmoordyn.body_get_type(instance)


def GetBodyState(instance):
    """ Get the body state

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance

    Returns:
    r: The position (a tuple with 3 components)
    u: The velocity (a tuple with 3 components)
    """
    import cmoordyn
    return cmoordyn.body_get_state(instance)


#                                  Rod.h
#  =============================================================================


def GetRodID(instance):
    """ Get the rod id

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance

    Returns:
    i: The rod id
    """
    import cmoordyn
    return cmoordyn.rod_get_id(instance)


def GetRodType(instance):
    """ Get the rod id

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance

    Returns:
    i: The rod type
    """
    import cmoordyn
    return cmoordyn.rod_get_type(instance)


def GetRodN(instance):
    """ Get the rod number of segments

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance

    Returns:
    n: The number of segments
    """
    import cmoordyn
    return cmoordyn.rod_get_n(instance)


def GetRodNodePos(instance, i):
    """ Get a rod node position

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance
    i (int): The node index

    Returns:
    r: The node pos
    """
    import cmoordyn
    return cmoordyn.rod_get_node_pos(instance, i)


#                                Connection.h
#  =============================================================================


def GetConnectID(instance):
    """ Get the connection id

    Parameters:
    instance (cmoordyn.MoorDynConnection): The connection instance

    Returns:
    i: The connection id
    """
    import cmoordyn
    return cmoordyn.conn_get_id(instance)


def GetConnectType(instance):
    """ Get the connection type

    Parameters:
    instance (cmoordyn.MoorDynConnection): The connection instance

    Returns:
    t: The connection type
    """
    import cmoordyn
    return cmoordyn.conn_get_type(instance)


def GetConnectPos(instance):
    """ Get the connection position

    Parameters:
    instance (cmoordyn.MoorDynConnection): The connection instance

    Returns:
    v: The connection position
    """
    import cmoordyn
    return cmoordyn.conn_get_pos(instance)


def GetConnectVel(instance):
    """ Get the connection velocity

    Parameters:
    instance (cmoordyn.MoorDynConnection): The connection instance

    Returns:
    v: The connection velocity
    """
    import cmoordyn
    return cmoordyn.conn_get_vel(instance)


def GetConnectForce(instance):
    """ Get the connection force

    Parameters:
    instance (cmoordyn.MoorDynConnection): The connection instance

    Returns:
    v: The connection force
    """
    import cmoordyn
    return cmoordyn.conn_get_force(instance)


def GetConnectNAttached(instance):
    """ Get the number of attached lines

    Parameters:
    instance (cmoordyn.MoorDynConnection): The connection instance

    Returns:
    n: The number of attached lines
    """
    import cmoordyn
    return cmoordyn.conn_get_nattached(instance)


def GetConnectAttached(instance, i):
    """ Get an attached line

    Parameters:
    instance (cmoordyn.MoorDynConnection): The connection instance
    i (int): The index of the attachment

    Returns:
    l: The attached line
    e: The endpoint
    """
    import cmoordyn
    return cmoordyn.conn_get_attached(instance, i)


#                                  Line.h
#  =============================================================================


def GetLineID(instance):
    """ Get the line id

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    i: The line id
    """
    import cmoordyn
    return cmoordyn.line_get_id(instance)


def GetLineN(instance):
    """ Get the line number of segments

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    n: The number of segments
    """
    import cmoordyn
    return cmoordyn.line_get_n(instance)


def GetLineUnstretchedLength(instance):
    """ Get the line unstretched length

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    l: The unstreched length
    """
    import cmoordyn
    return cmoordyn.line_get_ulen(instance)


def GetLineNodePos(instance, i):
    """ Get a line node position

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    r: The node position
    """
    import cmoordyn
    return cmoordyn.line_get_node_pos(instance, i)


def GetLineNodeTen(instance, i):
    """ Get a line node tension

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    t: The node tension
    """
    import cmoordyn
    return cmoordyn.line_get_node_force(instance, i)


def GetLineNodeCurv(instance, i):
    """ Get a line node curvature

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    r: The node curvature
    """
    import cmoordyn
    return cmoordyn.line_get_node_curv(instance, i)


def GetLineFairTen(instance):
    """ Get the tension magnitude at the fairlead of a line

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    t: The tension magnitude
    """
    import cmoordyn
    return cmoordyn.line_get_fairlead_tension(instance)


def GetLineMaxTen(instance):
    """ Get the line maximum tension magnitude

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    t: The tension magnitude
    """
    import cmoordyn
    return cmoordyn.line_get_max_tension(instance)



"""
Copyright (c) 2022, Jose Luis Cercos-Pita <jlc@core-marine.com>

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# This file is used to build self-contained Python wrappers, which does not
# depends on the prebuilt libmoordyn.so.2. This is ideal to be used within
# cibuildwheel, which is producing the artifacts that can be uploaded later to
# Pipy
# In general you do not want to use this script by your own. It is meant for the
# core developers to setup the pipy packages

import os
import shutil
from setuptools import setup, find_packages, Extension
import sysconfig


DESC = """Python version of MoorDyn, a lumped-mass mooring dynamics model
intended for coupling with floating structure codes.

This is not a wrapper of MoorDyn, but an stand-alone version. Thus you can
just install and use this package, no need of the MoorDyn library in your
system.

Please, visit https://moordyn-v2.readthedocs.io to learn more about this package,
and MoorDyn in general.

If you detect any problem, please feel free to report the issue on the GitHub
page:
https://github.com/mattEhall/MoorDyn

WARNING: This is a pre-release version of MoorDyn v2. The API may be changed
in the future, breaking your program
"""

# We better copy here the moordyn module
shutil.rmtree('moordyn', ignore_errors=True)
shutil.copytree(os.path.join('wrappers', 'python', 'moordyn'), 'moordyn')

# Collect the source code files
MOORDYN_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                            'source')
MOORDYN_SRCS = []
for f in os.listdir(MOORDYN_PATH):
    if not os.path.isfile(os.path.join(MOORDYN_PATH, f)):
        continue
    if not f.lower().endswith(".c") and not f.lower().endswith(".cpp"):
        continue
    MOORDYN_SRCS.append(os.path.join('source', f))
MOORDYN_SRCS.append(os.path.join('wrappers', 'python', 'cmoordyn.cpp'))

# Read the version from the CMakeLists.txt
version = ""
with open('CMakeLists.txt', 'r') as f:
    txt = f.read()
    for name in ('MAJOR', 'MINOR', 'PATCH'):
        prefix = 'set(MOORDYN_{}_VERSION '.format(name)
        subtxt = txt[txt.find(prefix) + len(prefix):]
        subtxt = subtxt[:subtxt.find(')\n')]
        version = version + subtxt + '.'
    version = version[:-1]

# Eigen needs at least C++ 14, and Moordyn itself uses C++ 17
extra_compile_args = ["-std=c++17"]

cmoordyn = Extension('cmoordyn',
                     sources=MOORDYN_SRCS,
                     language='c++',
                     extra_compile_args=extra_compile_args,
                     include_dirs=[MOORDYN_PATH, ],
                     define_macros=[('MoorDyn_EXPORTS', '1')],
                     )

setup(
    name='moordyn',
    version=version,
    author='Jose Luis Cercos-Pita',
    author_email='jlc@core-marine.com',
    url = 'https://github.com/mattEhall/MoorDyn',
    description='MoorDyn for Python',
    long_description = DESC,
    ext_modules = [cmoordyn],
    packages=find_packages(include=['moordyn', 'moordyn.*']),
    install_requires=[],
    setup_requires=['cython', 'pytest-runner'],
    tests_require=['pytest'],
)



class Entity:
    """Base class for all the other entities
    """
    def __init__(self,
                 section_name,
                 extra_header="",
                 field_names=[],
                 field_units=[]):
        """Constructor

        Parameters
        ----------
        section_name (str): The section name

        Keyword arguments
        -----------------
        extra_header (str): String to be added right after the header line.
                            Usually just the Intro section might accept that.
        field_names (list): The list of field names
        field_units (list): The list of field units
        """
        self.__name = section_name
        self.__extra_header = extra_header
        self.__field_names = field_names
        self.__field_units = field_units
        self.__values = None
        self.__field_widths = self.get_field_widths()

    def set_values(self, values):
        """Set the field values

        Parameters
        ----------
        values (list): The list of values
        """
        self.__values = values
        self.__field_widths = self.get_field_widths()

    def set_field_widths(self, widths):
        """Set the widths

        Parameters
        ----------
        widths (list): The list of strig width for each field
        """
        self.__field_widths = widths

    def get_field_widths(self):
        """Get the minimum width required to print the fields

        Returns
        -------
        list: The list of widths, which might not match the one set with
              set_field_widths() if more space is required for a field
        """
        widths = [len(field) + 1 for field in self.__field_names]
        for i, unit in enumerate(self.__field_units):
            widths[i] = max(widths[i], len(unit) + 1)
        if self.__values:
            for i in range(len(self.__values) - len(widths)):
                # For the cases without field names
                widths.append(0)
            for i, value in enumerate(self.__values):
                widths[i] = max(widths[i], len(str(value)) + 1)
        try:
            for i, w in enumerate(self.__field_widths):
                widths[i] = max(widths[i], w)
        except AttributeError:
            pass
        return widths

    def get_header(self):
        """Get the header string

        Returns
        -------
        str: The header string. This should be asked just for the first entity
        """
        hdr = (" " + self.__name + " ").center(80, "-") + "\n"
        if not hdr.startswith("---"):
            hdr = "---" + hdr
        hdr = hdr + self.__extra_header
        if not hdr.endswith("\n"):
            hdr = hdr + "\n"

        for i, name in enumerate(self.__field_names):
            hdr = hdr + name.ljust(self.__field_widths[i], " ")
        if not hdr.endswith("\n"):
            hdr = hdr + "\n"

        for i, unit in enumerate(self.__field_units):
            hdr = hdr + unit.ljust(self.__field_widths[i], " ")
        if not hdr.endswith("\n"):
            hdr = hdr + "\n"
        return hdr

    def get_values(self):
        """Get the string of values

        Returns
        -------
        str: The values string
        """
        values = ""
        for i, value in enumerate(self.__values):
            values = values + str(value).ljust(self.__field_widths[i], " ")
        if not values.endswith("\n"):
            values = values + "\n"
        return values


def to_list(v):
    try:
        return list(v)[:]
    except TypeError:
        return [v]


class PropsList:
    """The line connection
    """
    def __init__(self, l, sep):
        """Constructor

        Parameters
        ----------
        l (list): The list
        sep (str): Fields separator
        """
        self.__values = to_list(l)
        self.__sep = sep

    @property
    def values(self):
        return self.__values

    @values.setter
    def values(self, v):
        self.__values = to_list(v)

    @property
    def separator(self):
        return self.__end_point

    @separator.setter
    def separator(self, v):
        self.__separator = v

    def __str__(self):
        """Get the string

        Returns
        -------
        str: The string
        """
        out = ""
        for value in self.__values[:-1]:
            out = out + str(value) + self.__sep
        out = out + str(self.__values[-1])
        return out



from .Mooring import Mooring
from .Body import BodyConnection, Body
from .Connection import Connection
from .Failure import Failure
from .Line import LineMaterial, LineConnection, Line
from .Option import Option
from .Rod import RodMaterial, RodConnection, Rod
